import typer
from sentential.lib.drivers.local_lambda import LocalLambdaDriver
from sentential.lib.drivers.aws_lambda import AwsLambdaDriver
from sentential.lib.drivers.aws_api_gateway import AwsApiGatewayDriver
from sentential.lib.ontology import Ontology

destroy = typer.Typer()


@destroy.command()
def local():
    """destroy lambda deployment in aws"""
    local = LocalLambdaDriver(Ontology())
    local.destroy()


@destroy.command()
def aws():
    """destroy lambda deployment in aws"""
    ontology = Ontology()
    aws_function = AwsLambdaDriver(ontology)
    deployed = aws_function.deployed()
    aws_function.destroy(deployed)



from builtins import BaseException
from pydantic import BaseModel, ValidationError, Extra, Field
from typing import Any, List, Tuple, cast
from enum import Enum
import polars as pl
import ast

#
# User Driven Shapes
#

ALLOWED_LIST_TYPES = [List[str], List[int], List[float]]
ALLOWED_TYPES = [str, int, float] + ALLOWED_LIST_TYPES


class ShaperError(BaseException):
    pass


class Shaper(BaseModel):
    class Config:
        extra = Extra.forbid

    @classmethod
    def validate_field_value(cls, field: str, value: Any):
        try:
            validations = cls.constrained_validation_df({field: value})
        except KeyError as e:
            raise ShaperError(
                f"invalid key, valid options {list(cls.__fields__.keys())}"
            )

        validations = validations.filter(pl.col("field") == field)

        if len(validations) != 1:
            raise ShaperError(
                f"number of validations for {field} must be 1, found {len(validations)}"
            )

        validation = validations.row(0)
        (key, validation_error) = cast(Tuple[str, str], validation)

        if validation_error is not None:
            raise ValueError(validation_error)

    @classmethod
    def constrained_parse_obj(cls, data: dict):
        for (name, field) in cls.__fields__.items():
            # disallow types
            if field.outer_type_ not in ALLOWED_TYPES:
                raise ShaperError(f"disallowed type {field.outer_type_}")

        # If data is populating from ssm, it's always a string.
        # A list is stored in ssm as "[ 'some', 'list' ]"
        # Use knowledge that the field is a list, and ast.literal_eval
        # to bring it back to a python list, pydantic will coerce what's within
        for field in cls.__fields__.values():
            if field.outer_type_ in ALLOWED_LIST_TYPES:
                if field.name in data and type(data[field.name]) is str:
                    data[field.name] = ast.literal_eval(data[field.name])

        return cls(**data)

    @classmethod
    def type_constraints_df(cls):
        columns = [("field", pl.Utf8), ("type_check", pl.Utf8)]
        fields = [name for name in cls.__fields__.keys()]
        type_checks = [
            f"disallowed type {f.outer_type_}"
            if f.outer_type_ not in ALLOWED_TYPES
            else None
            for f in list(cls.__fields__.values())
        ]

        return pl.DataFrame(
            [
                fields,
                type_checks,
            ],
            columns=columns,
        )

    @classmethod
    def validation_df(cls, data: dict):
        columns = [("field", pl.Utf8), ("value_check", pl.Utf8)]
        fields = [name for name in cls.__fields__.keys()]
        try:
            cls.constrained_parse_obj(data)
            return pl.DataFrame([fields, [None for f in fields]], columns=columns)
        except ValidationError as e:
            locations = [map(str, e["loc"]) for e in e.errors()]
            locations = ["/".join(loc) for loc in locations]
            messages = [e["msg"] for e in e.errors()]
            return pl.DataFrame(
                [locations, messages],
                columns=columns,
            )

    @classmethod
    def constrained_validation_df(cls, data: dict):
        type_constraints = cls.type_constraints_df()
        validations = cls.validation_df(data)

        df = type_constraints.join(validations, on="field", how="outer")
        df = df.with_columns(
            [
                (pl.col("type_check").fill_null(pl.col("value_check"))).alias(
                    "validation"
                )
            ]
        )
        return df.select([pl.col("field"), pl.col("validation")])

    @classmethod
    def schema_df(cls):
        fields = list(cls.schema()["properties"].keys())
        properties = list(cls.schema()["properties"].values())
        defaults = [str(p["default"]) if "default" in p else None for p in properties]
        descriptions = [
            p["description"] if "description" in p else None for p in properties
        ]
        columns = [("field", pl.Utf8), ("default", pl.Utf8), ("description", pl.Utf8)]
        return pl.DataFrame(
            [
                fields,
                defaults,
                descriptions,
            ],
            columns=columns,
        )



import typer
from typing import List
from rich import print
from sentential.lib.ontology import Ontology

configs = typer.Typer()


@configs.command()
def read() -> None:
    f"""read provisioning configs for lambda"""
    print(Ontology().configs.read())


@configs.command()
def write(key: str, value: List[str]):
    f"""write provisioning configs for lambda"""
    Ontology().configs.write(key, value)


@configs.command()
def delete(key: str):
    f"""delete provisioning configs for lambda"""
    Ontology().configs.delete(key)


@configs.command()
def clear():
    f"""delete all provisioning configs for lambda"""
    Ontology().configs.clear()



import polars as pl
from typing import List, Type
from types import SimpleNamespace
from rich.table import Table, box
from builtins import ValueError
from sentential.lib.exceptions import StoreError
from sentential.lib.clients import clients
from sentential.lib.context import Context
from sentential.support.shaper import Shaper, ShaperError


class Common:
    path: str
    context: Context

    def _fetch(self):
        return clients.ssm.get_parameters_by_path(
            Path=self.path,
            Recursive=True,
            WithDecryption=True,
        )["Parameters"]

    def as_dict(self):
        return {p["Name"].replace(self.path, ""): p["Value"] for p in self._fetch()}

    def as_df(self):
        data = self.as_dict()
        data = [list(data.keys()), list(data.values())]
        return pl.DataFrame(data, columns=[("field", pl.Utf8), ("value", pl.Utf8)])

    def delete(self, key: str):
        try:
            return clients.ssm.delete_parameter(Name=f"{self.path}{key}")
        except clients.ssm.exceptions.ParameterNotFound:
            raise StoreError(f"no key '{key}' persisted")

    def clear(self):
        for (key, value) in self.as_dict().items():
            self.delete(key)


class GenericStore(Common):
    def __init__(self, context: Context, prefix: str) -> None:
        super().__init__()
        self.context = context
        self.path = (
            f"/{self.context.partition}/{self.context.repository_name}/{prefix}/"
        )

    @property
    def parameters(self) -> SimpleNamespace:
        return SimpleNamespace(**self.as_dict())

    def read(self) -> Table:
        data = self.as_dict()
        table = Table("field", "value", box=box.SIMPLE)
        for (key, value) in data.items():
            table.add_row(key, value)
        return table

    def write(self, key: str, value: List[str]):
        # Typer doesn't support Union[str, List[str]], understandably
        # So desired behavior is implemented here, to the insult of typing

        delimiter = ","
        value = delimiter.join(value)  # type: ignore

        return clients.ssm.put_parameter(
            Name=f"{self.path}{key}",
            Value=str(value),
            Type="SecureString",
            Overwrite=True,
            Tier="Standard",
            KeyId=self.context.kms_key_id,
        )

    def export_defaults(self) -> None:
        return None


class ModeledStore(Common):
    def __init__(self, context: Context, prefix: str, model: Type[Shaper]) -> None:
        super().__init__()
        self.context = context
        self.model = model
        self.path = (
            f"/{self.context.partition}/{self.context.repository_name}/{prefix}/"
        )

    def export_defaults(self) -> None:
        if self.model:
            current_state = self.as_dict()
            for (name, field) in self.model.__fields__.items():
                if field.name not in current_state.keys():
                    if hasattr(field, "default") and field.default != None:
                        self.write(field.name, field.default)

    @property
    def parameters(self) -> Shaper:
        return self.model.constrained_parse_obj(self.as_dict())

    def read(self) -> Table:
        data = self.as_df()
        opts = {"on": "field", "how": "outer"}
        schema = self.model.schema_df()
        validation = self.model.constrained_validation_df(self.as_dict())
        df = schema.join(data, **opts).join(validation, **opts)
        df = df.with_columns([(pl.col("value").fill_null(pl.col("default")))])
        df = df.select(
            [
                pl.col("field"),
                pl.col("value"),
                pl.col("validation"),
                pl.col("description"),
            ]
        )

        table = Table(*df.columns, box=box.SIMPLE)
        for row in df.rows():
            table.add_row(*[str(value) for value in row])
        return table

    def write(self, key: str, value: List[str]):

        # Typer doesn't support Union[str, List[str]], understandably
        # So desired behavior is implemented here, to the insult of typing
        if len(value) == 1:
            value = value[0]  # type: ignore

        try:
            self.model.validate_field_value(key, value)
        except ValueError as e:
            raise StoreError(e)
        except ShaperError as e:
            raise StoreError(e)

        return clients.ssm.put_parameter(
            Name=f"{self.path}{key}",
            Value=str(value),
            Type="SecureString",
            Overwrite=True,
            Tier="Standard",
            KeyId=self.context.kms_key_id,
        )



import typer
from typing import List
from rich import print
from sentential.lib.ontology import Ontology

args = typer.Typer()


@args.command()
def read() -> None:
    f"""read build arguments for lambda"""
    print(Ontology().args.read())


@args.command()
def write(key: str, value: List[str]):
    f"""write build arguments for lambda"""
    Ontology().args.write(key, value)


@args.command()
def delete(key: str):
    f"""delete build arguments for lambda"""
    Ontology().args.delete(key)


@args.command()
def clear():
    f"""delete all build arguments for lambda"""
    Ontology().args.clear()



import os
import sys
import importlib
from typing import Union
from sentential.lib.context import Context
from sentential.lib.store import GenericStore, ModeledStore
from sentential.lib.shapes import Provision as Model


def reload_shapes():
    if os.getcwd() not in sys.path:
        sys.path.append(os.getcwd())
    import shapes

    importlib.reload(shapes)


class Ontology:
    def __init__(self) -> None:
        pass

    @property
    def context(cls) -> Context:
        return Context()

    @property
    def args(cls) -> Union[GenericStore, ModeledStore]:
        try:
            reload_shapes()
            from shapes import Args as Model  # type: ignore

            return ModeledStore(cls.context, "arg", Model)
        except:
            return GenericStore(cls.context, "arg")

    @property
    def envs(cls) -> Union[GenericStore, ModeledStore]:
        try:
            reload_shapes()
            from shapes import Envs as Model  # type: ignore

            return ModeledStore(cls.context, "env", Model)
        except:
            return GenericStore(cls.context, "env")

    @property
    def configs(cls) -> ModeledStore:
        return ModeledStore(cls.context, "provision", Model)



import boto3
from python_on_whales import docker
from sentential.lib.exceptions import AWS_EXCEPTIONS


class Clients:
    def __init__(self) -> None:
        self.lmb = boto3.client("lambda")
        self.iam = boto3.client("iam")
        self.sts = boto3.client("sts")
        self.ecr = boto3.client("ecr")
        self.ssm = boto3.client("ssm")
        self.api_gw = boto3.client("apigatewayv2")
        self.cloudwatch = boto3.client("logs")
        self.docker = docker


try:
    clients = Clients()
except AWS_EXCEPTIONS as e:
    print(f"AWS: {e}")
    exit(1)



import botocore.exceptions
from python_on_whales.exceptions import DockerException


def gather_aws_exceptions() -> tuple:
    exceptions = []
    for key, value in sorted(botocore.exceptions.__dict__.items()):
        if isinstance(value, type):
            exceptions.append(getattr(botocore.exceptions, key))
    return tuple(exceptions)


AWS_EXCEPTIONS = gather_aws_exceptions()


class SntlException(BaseException):
    pass


class AwsDriverError(SntlException):
    pass


class LocalDriverError(SntlException):
    pass


class ContextError(SntlException):
    pass


class StoreError(SntlException):
    pass


class JoineryError(SntlException):
    pass


class ApiGatewayResourceNotFound(SntlException):
    pass



from abc import ABC, abstractmethod


class LambdaDriver(ABC):
    @abstractmethod
    def deployed(cls):
        ...

    @abstractmethod
    def images(self):
        ...

    @abstractmethod
    def image(self, version: str):
        ...

    @abstractmethod
    def deploy(self):
        ...

    @abstractmethod
    def destroy(self):
        ...

    @abstractmethod
    def logs(self, follow: bool):
        ...

    @abstractmethod
    def invoke(self, payload: str):
        ...



import os
from tempfile import TemporaryDirectory
from typing import Dict, List
from sentential.lib.exceptions import LocalDriverError
from sentential.lib.clients import clients
from sentential.lib.drivers.spec import LambdaDriver
from sentential.lib.ontology import Ontology
from sentential.lib.shapes import Image, Function
from sentential.lib.template import Policy
from python_on_whales.components.image.cli_wrapper import Image as DriverImage


#
# NOTE: Docker images locally are primary key'd (conceptually) off of their id, this is normalized by the Image type
#


class LocalLambdaDriver(LambdaDriver):
    def __init__(self, ontology: Ontology) -> None:
        self.ontology = ontology

    def build(self, version: str) -> Image:
        self.ontology.args.export_defaults()

        built_image = clients.docker.build(
            self.ontology.context.path.root,
            load=True,
            tags=[f"{self.ontology.context.repository_name}:{version}"],
            build_args=self.ontology.args.as_dict(),
        )

        if isinstance(built_image, DriverImage):
            return self._image_where_id(built_image.id)
        else:
            raise LocalDriverError("build returned unexpected type")

    def pull(self, image: Image) -> List[str]:
        tags_pulled = []
        for tag in image.tags:
            if self.ontology.context.repository_url in tag:
                clients.docker.pull(tag)
                tags_pulled.append(tag)
        return tags_pulled

    def publish(self, source_version: str, destination_version: str) -> Image:
        image = self.image(source_version)
        repo_url = self.ontology.context.repository_url
        shipping_tag = f"{repo_url}:{destination_version}"
        clients.docker.tag(image.id, shipping_tag)
        clients.docker.push(shipping_tag)
        print(f"published {image.id} as {shipping_tag}")
        return self.image(destination_version)

    def images(self) -> List[Image]:
        images = []
        for id, image in self._docker_data().items():
            images.append(
                Image(
                    id=id,
                    digest=image["digest"],
                    tags=image["tags"],
                    versions=image["versions"],
                )  # arch=image.architecture
            )
        return images

    def image(self, version: str) -> Image:
        for image in self.images():
            if version in image.versions:
                return image
        raise LocalDriverError(f"no image found where version {version}")

    def deployed(self) -> Function:
        # TODO: "sentential" container name is not a good enough matching mechanism
        running = [c for c in clients.docker.ps() if c.name == "sentential"]
        public_url = [c for c in clients.docker.ps() if c.name == "sentential-gw"]
        if running:
            container = running[0]
            running_image = clients.docker.image.inspect(container.image)
            image = self._image_where_id(running_image.id)

            if public_url:
                public_url = "http://localhost:8081"
            else:
                public_url = None

            return Function(
                image=image,
                region=self.ontology.context.region,
                name="local",
                arn="local",
                role_name="local",
                role_arn="local",
                public_url=public_url,
                web_console_url=None,
            )
        raise LocalDriverError(f"no image found with container name sentential")

    def deploy(self, image: Image, public_url: bool) -> str:
        self.ontology.envs.export_defaults()
        self.destroy()
        clients.docker.network.create("sentential-bridge")
        credentials = self._get_federation_token()
        default_env = {
            "AWS_REGION": self.ontology.context.region,
            "PARTITION": self.ontology.envs.path,
        }

        clients.docker.run(
            image.id,
            name="sentential",
            hostname="sentential",
            networks=["sentential-bridge"],
            detach=True,
            remove=False,
            publish=[("9000", "8080")],
            envs={**default_env, **credentials},
        )

        if public_url:
            clients.docker.run(
                "ghcr.io/wheegee/sentential-gw:latest",
                name="sentential-gw",
                hostname="sentential-gw",
                networks=["sentential-bridge"],
                detach=True,
                remove=False,
                publish=[("8081", "8081")],
                envs={"LAMBDA_ENDPOINT": "http://sentential:8080"},
            )

        if public_url:
            return "http://localhost:8081"
        else:
            return "http://localhost:9000"

    def destroy(self):
        clients.docker.remove(["sentential"], force=True, volumes=True)
        clients.docker.remove(["sentential-gw"], force=True, volumes=True)
        try:
            clients.docker.network.remove(["sentential-bridge"])
        except:
            pass

    def logs(self, follow: bool = False):
        cmd = ["docker", "logs", "sentential"]
        if follow:
            cmd.append("--follow")
        os.system(" ".join(cmd))

    def invoke(self, payload: str):
        # TODO: move away from shell-out here
        with TemporaryDirectory() as tmp:
            cmd = [
                "aws",
                "lambda",
                "invoke",
                "--function-name",
                "function",
                "--endpoint",
                "http://localhost:9000",
                "--log-type",
                "Tail",
                "--invocation-type",
                "RequestResponse",
                "--cli-binary-format",
                "raw-in-base64-out",
                "--payload",
                f"'{payload}'",
                f"{tmp}/output",
            ]
            os.system(" ".join(cmd))
            os.system(f"cat {tmp}/output")

    def _get_federation_token(self):
        policy_json = Policy(self.ontology).render()
        token = clients.sts.get_federation_token(
            Name=f"{self.ontology.context.repository_name}-spec-policy",
            Policy=policy_json,
        )["Credentials"]

        return {
            "AWS_ACCESS_KEY_ID": token["AccessKeyId"],
            "AWS_SECRET_ACCESS_KEY": token["SecretAccessKey"],
            "AWS_SESSION_TOKEN": token["SessionToken"],
        }

    def _docker_data(self) -> Dict:
        docker_data = {}
        repo_name = self.ontology.context.repository_name
        for image in clients.docker.images():
            # strip versions
            repo_names_w_url = [tag.split(":")[0] for tag in image.repo_tags]
            # strip urls
            repo_names = [tag.split("/")[-1] for tag in repo_names_w_url]
            # match against known repo name
            match = any([repo_name == name for name in repo_names])

            if match:
                digest = None
                for proposed_digest in image.repo_digests:
                    proposed_digest = proposed_digest.split("@")[-1]

                    # safety: if assumption that image id and image digest are always tightly coupled is untrue, raise plz
                    if digest is not None:
                        if digest != proposed_digest:
                            raise LocalDriverError(
                                "image id and image digest not tightly coupled"
                            )

                    digest = proposed_digest

                versions = []
                for tag in image.repo_tags:
                    versions.append(tag.split(":")[-1])

                docker_data[image.id] = {
                    "id": image.id,
                    "digest": digest,
                    "tags": image.repo_tags,
                    "versions": versions,
                }

        return docker_data

    def _image_where_id(self, id: str) -> Image:
        for image in self.images():
            if id == image.id:
                return image
        raise LocalDriverError(f"no image found with for id {id}")



import typer
from typing import List
from rich import print
from sentential.lib.ontology import Ontology

envs = typer.Typer()


@envs.command()
def read() -> None:
    f"""read environment variables for lambda"""
    print(Ontology().envs.read())


@envs.command()
def write(key: str, value: List[str]):
    f"""write environment variables for lambda"""
    Ontology().envs.write(key, value)


@envs.command()
def delete(key: str):
    f"""delete environment variables for lambda"""
    Ontology().envs.delete(key)


@envs.command()
def clear():
    f"""delete all environment variables for lambda"""
    Ontology().envs.clear()



from operator import concat
from typing import Dict, List
from rich.table import Table, box
from sentential.lib.drivers.aws_lambda import AwsLambdaDriver
from sentential.lib.drivers.local_lambda import LocalLambdaDriver
from sentential.lib.ontology import Ontology
from sentential.lib.shapes import Image, ImageView


class Joinery:
    def __init__(self, ontology: Ontology) -> None:
        self.ontology = ontology
        self.local = LocalLambdaDriver(self.ontology)
        self.aws = AwsLambdaDriver(self.ontology)
        try:
            self.local_deployment = self.local.deployed()
        except:
            self.local_deployment = None

        try:
            self.aws_deployment = self.aws.deployed()
        except:
            self.aws_deployment = None

    def group_by_id(self) -> Dict[str, List[Image]]:
        local_images = [image for image in self.local.images()]
        aws_images = [image for image in self.aws.images()]
        by_id = {}
        for image in local_images + aws_images:
            if image.id in by_id:
                by_id[image.id].append(image)
            else:
                by_id[image.id] = [image]
        return by_id

    def merge_on_id(self) -> List[ImageView]:
        merged = []

        for id, images in self.group_by_id().items():
            digest = None
            tags = []
            versions = []
            href = []
            for image in images:
                if image.digest is not None:
                    digest = image.digest

                if image.tags:
                    tags = concat(tags, image.tags)

                if image.versions:
                    versions = concat(versions, image.versions)

                if self.local_deployment:
                    if self.local_deployment.image.id == image.id:
                        public_url = self.local_deployment.public_url
                        if public_url:
                            href.append(f"[link={public_url}]local_url[/link]")

                if self.aws_deployment:
                    if self.aws_deployment.image.id == image.id:
                        web_console_url = self.aws_deployment.web_console_url
                        public_url = self.aws_deployment.public_url
                        if web_console_url:
                            href.append(f"[link={web_console_url}]aws_console[/link]")
                        if public_url:
                            href.append(f"[link={public_url}]public_url[/link]")

            merged.append(
                ImageView(
                    id=id,
                    digest=digest,
                    tags=list(tags),
                    versions=list(versions),
                    href=href,
                )
            )

        return merged

    def list(self, drop: List[str] = []) -> Table:
        columns = list((ImageView.__fields__.keys()))

        for header in drop:
            if header in columns:
                columns.remove(header)

        data = [list(i.dict(exclude=set(drop)).values()) for i in self.merge_on_id()]
        table = Table(box=box.SIMPLE, *columns)
        for row in data:
            table.add_row(*[str(value) for value in row])
        return table



import json
import os
from time import sleep
from typing import Dict, List, Optional, cast
from sentential.lib.exceptions import AwsDriverError
from sentential.lib.drivers.spec import LambdaDriver
from sentential.lib.ontology import Ontology
from sentential.lib.shapes import (
    LAMBDA_ROLE_POLICY_JSON,
    AwsFunctionPublicUrl,
    Image,
    AwsFunction,
    Function,
    Provision,
)
from sentential.lib.clients import clients
from sentential.lib.template import Policy

# TODO: figure out how to get boto3 typeshed to work, so functions can return types instead of Dict

#
# NOTE: Docker images in ECR are primary key'd (conceptually) off of their digest, this is normalized by the Image type
#


class AwsLambdaDriver(LambdaDriver):
    def __init__(self, ontology: Ontology) -> None:
        self.ontology = ontology
        self.context = self.ontology.context
        self.partition = self.context.partition
        self.region = self.context.region
        self.account_id = self.context.account_id
        self.repo_name = self.context.repository_name
        self.repo_url = self.context.repository_url
        self.envs = self.ontology.envs
        self.resource_name = f"{self.partition}-{self.region}-{self.repo_name}"
        self.policy_arn = f"arn:aws:iam::{self.account_id}:policy/{self.resource_name}"
        self.web_console_url = f"https://{self.region}.console.aws.amazon.com/lambda/home#/functions/{self.resource_name}"

    @property
    def provision(self) -> Provision:
        # there must be a better way to do polymorphic type stuff...
        return cast(Provision, self.ontology.configs.parameters)

    def deployed(self) -> Function:
        try:
            response = clients.lmb.get_function(FunctionName=self.resource_name)
            function = AwsFunction(**response)
        except clients.lmb.exceptions.ResourceNotFoundException:
            raise AwsDriverError(
                f"could not find aws deployed function for {self.resource_name}"
            )

        try:
            response = clients.lmb.get_function_url_config(
                FunctionName=function.Configuration.FunctionName
            )
            function_public_url = AwsFunctionPublicUrl(**response)
        except clients.lmb.exceptions.ResourceNotFoundException:
            function_public_url = None

        if function_public_url:
            public_url = function_public_url.FunctionUrl
        else:
            public_url = None

        digest = function.Code.ResolvedImageUri.split("@")[-1]
        image = self._image_where_digest(digest)

        return Function(
            image=image,
            name=function.Configuration.FunctionName,
            arn=function.Configuration.FunctionArn,
            role_arn=function.Configuration.Role,
            role_name=function.Configuration.Role.split("/")[-1],
            region=self.context.region,
            web_console_url=self.web_console_url,
            public_url=public_url,
        )

    def images(self) -> List[Image]:
        images = []
        for digest, image in self._ecr_data().items():
            images.append(
                Image(
                    id=image["id"],
                    digest=digest,
                    tags=image["tags"],
                    versions=image["versions"],
                )
            )

        return images

    def image(self, version: str) -> Image:
        for image in self.images():
            if version in image.versions:
                return image
        raise AwsDriverError(f"no image found where version is {version}")

    def deploy(self, image: Image, public_url: bool) -> Function:
        self.ontology.envs.export_defaults()
        self._put_role()
        clients.iam.attach_role_policy(
            RoleName=self._put_role()["Role"]["RoleName"],
            PolicyArn=self._put_policy()["Policy"]["Arn"],
        )

        self._put_lambda(image)

        if public_url:
            self._put_url()

        return self.deployed()

    def destroy(self, function: Function) -> None:
        try:
            clients.lmb.delete_function_url_config(FunctionName=function.name)
        except clients.lmb.exceptions.ResourceNotFoundException:
            pass

        try:
            clients.lmb.delete_function(FunctionName=function.name)
        except clients.lmb.exceptions.ResourceNotFoundException:
            pass

        try:
            clients.iam.detach_role_policy(
                PolicyArn=self.policy_arn, RoleName=function.role_name
            )

            policy_versions = clients.iam.list_policy_versions(
                PolicyArn=self.policy_arn
            )["Versions"]
            for policy_version in policy_versions:
                if not policy_version["IsDefaultVersion"]:
                    clients.iam.delete_policy_version(
                        PolicyArn=self.policy_arn, VersionId=policy_version["VersionId"]
                    )

            clients.iam.delete_policy(PolicyArn=self.policy_arn)
        except clients.iam.exceptions.NoSuchEntityException:
            pass

        try:
            clients.iam.delete_role(RoleName=function.role_name)
        except clients.iam.exceptions.NoSuchEntityException:
            pass

    def logs(self, follow: bool = False) -> None:
        cmd = ["aws", "logs", "tail", f"/aws/lambda/{self.resource_name}"]
        if follow:
            cmd.append("--follow")
        os.system(" ".join(cmd))

    def invoke(self, payload: str) -> None:
        raise AwsDriverError("invoke is not yet implemented")

    def _put_role(self, tags: Optional[Dict[str, str]] = None) -> Dict:
        role_name = self.resource_name
        try:
            clients.iam.create_role(
                RoleName=role_name,
                AssumeRolePolicyDocument=LAMBDA_ROLE_POLICY_JSON,
            )

            clients.iam.get_waiter("role_exists").wait(RoleName=role_name)

        except clients.iam.exceptions.EntityAlreadyExistsException:
            clients.iam.update_assume_role_policy(
                RoleName=role_name,
                PolicyDocument=LAMBDA_ROLE_POLICY_JSON,
            )

        clients.iam.get_waiter("role_exists").wait(RoleName=role_name)

        if tags:
            clients.iam.tag_role(
                RoleName=role_name,
                Tags=[{"Key": key, "Value": value} for (key, value) in tags.items()],
            )

        return clients.iam.get_role(RoleName=role_name)

    def _put_policy(self, tags: Optional[Dict[str, str]] = None) -> Dict:
        policy_json = Policy(self.ontology).render()
        policy_name = self.resource_name
        policy_arn = self.policy_arn

        try:
            policy = clients.iam.create_policy(
                PolicyName=policy_name,
                PolicyDocument=policy_json,
            )

        except clients.iam.exceptions.EntityAlreadyExistsException:
            policy = clients.iam.get_policy(PolicyArn=policy_arn)

            versions = clients.iam.list_policy_versions(PolicyArn=policy_arn)[
                "Versions"
            ]

            if len(versions) >= 5:
                clients.iam.delete_policy_version(
                    PolicyArn=policy_arn, VersionId=versions[-1]["VersionId"]
                )

            clients.iam.create_policy_version(
                PolicyArn=policy_arn,
                PolicyDocument=policy_json,
                SetAsDefault=True,
            )

        if tags:
            clients.iam.tag_policy(
                PolicyName=policy.name,
                Tags=[{"Key": key, "Value": value} for (key, value) in tags.items()],
            )

        clients.iam.get_waiter("policy_exists").wait(PolicyArn=policy_arn)
        return policy

    def _put_lambda(self, image: Image, tags: Optional[Dict[str, str]] = None) -> Dict:
        role_name = self.resource_name
        function_name = self.resource_name
        role_arn = clients.iam.get_role(RoleName=role_name)["Role"]["Arn"]
        image_uri = f"{self.repo_url}:{image.versions[0]}"  # TODO: do we want to deploy latest version on image, or version declared?
        envs_path = self.envs.path
        sleep(10)
        try:
            function = clients.lmb.create_function(
                FunctionName=function_name,
                Role=role_arn,
                PackageType="Image",
                Code={"ImageUri": image_uri},
                Description=f"sententially deployed {image_uri}",
                Environment={"Variables": {"PARTITION": envs_path}},
                # Architectures=[self.image.arch], # TODO: figure out wtf is going on with fetching arch.
                EphemeralStorage={"Size": self.provision.storage},
                MemorySize=self.provision.memory,
                Timeout=self.provision.timeout,
                VpcConfig={
                    "SubnetIds": self.provision.subnet_ids,
                    "SecurityGroupIds": self.provision.security_group_ids,
                },
            )

            clients.lmb.add_permission(
                FunctionName=function_name,
                StatementId="FunctionURLAllowPublicAccess",
                Action="lambda:InvokeFunctionUrl",
                Principal="*",
                FunctionUrlAuthType=self.provision.auth_type,
            )

            return function
        except clients.lmb.exceptions.ResourceConflictException:
            function = clients.lmb.update_function_configuration(
                FunctionName=function_name,
                Role=role_arn,
                Description=f"sententially deployed {image_uri}",
                Environment={"Variables": {"PARTITION": envs_path}},
                EphemeralStorage={"Size": self.provision.storage},
                MemorySize=self.provision.memory,
                Timeout=self.provision.timeout,
                VpcConfig={
                    "SubnetIds": self.provision.subnet_ids,
                    "SecurityGroupIds": self.provision.security_group_ids,
                },
            )

            clients.lmb.get_waiter("function_updated_v2").wait(
                FunctionName=function_name
            )

            clients.lmb.update_function_code(
                FunctionName=function_name,
                ImageUri=image_uri,
                Publish=True,
            )

            try:
                clients.lmb.remove_permission(
                    FunctionName=function_name,
                    StatementId="FunctionURLAllowPublicAccess",
                )
            except clients.lmb.exceptions.ResourceNotFoundException:
                pass

            clients.lmb.add_permission(
                FunctionName=function_name,
                StatementId="FunctionURLAllowPublicAccess",
                Action="lambda:InvokeFunctionUrl",
                Principal="*",
                FunctionUrlAuthType=self.provision.auth_type,
            )

            if tags:
                clients.lmb.tag_resource(Resource=function.arn, Tags=tags)

            return function

    def _put_url(self) -> Dict:
        function_name = self.resource_name
        config = {
            "FunctionName": function_name,
            "AuthType": self.provision.auth_type,
            "Cors": {
                "AllowHeaders": self.provision.allow_headers,
                "AllowMethods": self.provision.allow_methods,
                "AllowOrigins": self.provision.allow_origins,
                "ExposeHeaders": self.provision.expose_headers,
            },
        }

        try:
            clients.lmb.create_function_url_config(**config)
        except clients.lmb.exceptions.ResourceConflictException:
            clients.lmb.update_function_url_config(**config)

        return clients.lmb.get_function_url_config(FunctionName=function_name)

    def _ecr_data(self) -> Dict:
        ecr_data = {}
        describe_images = clients.ecr.describe_images(repositoryName=self.repo_name)[
            "imageDetails"
        ]

        if len(describe_images) == 0:
            return {}

        image_digests = [
            {"imageDigest": image["imageDigest"]} for image in describe_images
        ]

        batch_get_images = clients.ecr.batch_get_image(
            repositoryName=self.repo_name, imageIds=image_digests
        )["images"]

        for image in describe_images:
            if "imageTags" in image:
                versions = image["imageTags"]
                tags = [f"{self.repo_url}:{tag}" for tag in image["imageTags"]]
            else:
                versions = []
                tags = []

            ecr_data[image["imageDigest"]] = {"versions": versions, "tags": tags}

        for image in batch_get_images:
            image_digest = image["imageId"]["imageDigest"]
            image_manifest = json.loads(image["imageManifest"])
            image_id = image_manifest["config"]["digest"]

            # safety: if assumption that image id and image digest are always tightly coupled is untrue, raise plz
            if "id" in ecr_data[image_digest]:
                if ecr_data[image_digest]["id"] != image_id:
                    raise AwsDriverError(
                        "image id and image digest not tightly coupled"
                    )

            ecr_data[image_digest]["id"] = image_id

        return ecr_data

    def _image_where_digest(self, digest: str) -> Image:
        for image in self.images():
            if digest == image.digest:
                return image
        raise AwsDriverError(f"no image found where digest is {digest}")



import re
from typing import List
import semantic_version as semver
from distutils.version import LooseVersion

from sentential.lib.shapes import Image

# pyright: reportInvalidStringEscapeSequence=false
SEMVER_REGEX = "^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"


class SemVer:
    def __init__(self, images: List[Image]) -> None:
        self.images = images

    @property
    def versions(self) -> List[str]:
        images = self.images
        versions = []
        for image in images:
            for version in image.versions:
                versions.append(version)
        return list(set(versions))

    @property
    def semver(self) -> List[str]:
        matcher = re.compile(SEMVER_REGEX)
        versions = self.versions
        versions = [version for version in versions if matcher.match(version)]
        versions = sorted(versions, key=lambda v: LooseVersion(v))
        return versions

    @property
    def latest(self) -> str:
        if self.semver:
            return self.semver[-1]
        else:
            return "0.0.0"

    def next(self, major=False, minor=False) -> str:
        latest = semver.Version(self.latest)
        if major:
            return str(latest.next_major())
        if minor:
            return str(latest.next_minor())
        else:
            return str(latest.next_patch())



import typer
from sentential.lib.drivers.aws_lambda import AwsLambdaDriver
from sentential.lib.drivers.local_lambda import LocalLambdaDriver
from sentential.lib.ontology import Ontology

logs = typer.Typer()


@logs.command()
def local(follow: bool = typer.Option(False)):
    """dump running container logs"""
    local = LocalLambdaDriver(Ontology())
    local.logs(follow)


@logs.command()
def aws(follow: bool = typer.Option(False)):
    """dump running container logs"""
    aws = AwsLambdaDriver(Ontology())
    aws.logs(follow)



import typer
from sentential.lib.drivers.aws_lambda import AwsLambdaDriver
from sentential.lib.drivers.local_lambda import LocalLambdaDriver
from sentential.lib.ontology import Ontology
from sentential.lib.shapes import CURRENT_WORKING_IMAGE_TAG

deploy = typer.Typer()


@deploy.command()
def local(
    version: str = typer.Argument(CURRENT_WORKING_IMAGE_TAG, envvar="VERSION"),
    public_url: bool = typer.Option(False),
):
    """build and deploy local lambda container"""
    local = LocalLambdaDriver(Ontology())
    aws = AwsLambdaDriver(Ontology())
    try:
        image = local.image(version)
    except:
        image = aws.image(version)
        local.pull(image)

    print(local.deploy(image, public_url))


@deploy.command()
def aws(
    version: str = typer.Argument(..., envvar="VERSION"),
    public_url: bool = typer.Option(default=False),
):
    """deploy lambda image to aws"""
    ontology = Ontology()
    aws_lambda = AwsLambdaDriver(ontology)
    image = aws_lambda.image(version)
    function = aws_lambda.deploy(image, public_url)
    if function.public_url:
        print(function.public_url)



from sentential.support.shaper import Shaper, List, Field

"""
# Define, describe, and validate environment variables and build arguments.

# See documentation for general usage.

# Example:
class Env(Shaper):
    db_url: str
    db_port: int = 3306
    db_username: str = "postgres"
    db_password: str

#
# Compatible Classes
#

class Arg(Shaper):
    pass

class Env(Shaper):
    pass
"""



from shutil import copy
from os import makedirs
from pathlib import PosixPath
from os.path import dirname, abspath, exists
from jinja2 import Environment, FileSystemLoader, Template
from sentential.lib.shapes import derive_paths
from sentential.lib.ontology import Ontology

PACKAGE_PATH = PosixPath(dirname(abspath(__file__))).parent


class Init:
    def __init__(self, repository_name: str, runtime: str) -> None:
        self.repository_name = repository_name
        self.runtime = runtime
        self.path = derive_paths()
        self.jinja = Environment(loader=FileSystemLoader(f"{PACKAGE_PATH}/templates"))

    def scaffold(self) -> None:
        if not exists(self.path.src):
            makedirs(self.path.src)
        self._write(self.jinja.get_template("Dockerfile"), self.path.dockerfile)
        copy(f"{PACKAGE_PATH}/templates/policy.json", self.path.policy)
        copy(f"{PACKAGE_PATH}/templates/shapes.py", self.path.shapes)

    def _write(self, template: Template, write_to: PosixPath) -> PosixPath:
        if not exists(write_to):
            with open(write_to, "w+") as f:
                f.writelines(
                    template.render(
                        repository_name=self.repository_name,
                        runtime=self.runtime,
                        paths=self.path,
                    )
                )

        return write_to


class Policy:
    def __init__(self, ontology: Ontology) -> None:
        self.ontology = ontology
        self.jinja = Environment(loader=FileSystemLoader("."))

    def render(self) -> str:
        template = self.jinja.get_template("policy.json")
        return template.render(
            context=self.ontology.context, env=self.ontology.envs.parameters
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sentential',
 'sentential.cli',
 'sentential.lib',
 'sentential.lib.drivers',
 'sentential.support',
 'sentential.templates']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'boto3>=1.24.31,<2.0.0',
 'flaky>=3.7.0,<4.0.0',
 'furl>=2.1.3,<3.0.0',
 'polars>=0.14.2,<0.15.0',
 'pydantic>=1.9.1,<2.0.0',
 'python-on-whales>=0.49,<0.51',
 'rich>=12.5.1,<13.0.0',
 'semantic-version>=2.10.0,<3.0.0',
 'tabulate>=0.8.10,<0.9.0',
 'typer>=0.6.1,<0.7.0',
 'types-boto3>=1.0.2,<2.0.0']

entry_points = \
{'console_scripts': ['sntl = sentential.sntl:main']}

setup_kwargs = {
    'name': 'sentential',
    'version': '0.5.0',
    'description': 'because lambdas are good',
    'long_description': 'None',
    'author': 'Brendan Keane',
    'author_email': 'btkeane@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import boto3
from os import getenv, environ
from sentential.lib.exceptions import ContextError
from sentential.lib.clients import clients
from sentential.lib.shapes import derive_paths, Paths, AWSCallerIdentity


class Context:
    @property
    def repository_name(self) -> str:
        try:
            repo = None
            with open("./Dockerfile") as file:
                for line in file.readlines():
                    if "FROM runtime AS" in line:
                        repo = line.split("AS")[1].strip()
                if repo is not None:
                    return repo
                else:
                    raise ContextError("Dockerfile not formed for sentential")
        except IOError:
            raise ContextError("no Dockerfile present, run `sntl init` first")

    @property
    def kms_key_alias(self) -> str:
        return getenv("AWS_KMS_KEY_ALIAS", default="aws/ssm")

    @property
    def caller_identity(self) -> AWSCallerIdentity:
        response = clients.sts.get_caller_identity()
        return AWSCallerIdentity(**response)

    @property
    def partition(self) -> str:
        if "PARTITION" in environ:
            return str(getenv("PARTITION"))
        else:
            user_id = self.caller_identity.UserId
            if ":" in user_id:
                # The ID before ':' in an assumed role seems to remain constant... time will tell.
                return str(user_id.split(":")[0])
            else:
                return user_id

    @property
    def region(self) -> str:
        return str(boto3.Session().region_name)

    @property
    def path(self) -> Paths:
        return derive_paths()

    @property
    def account_id(self) -> str:
        return self.caller_identity.Account

    @property
    def kms_key_id(self) -> str:
        # TODO: if region has not yet written an ssm param with the default key, the kms key will not yet exist \o/. Implement initialization of default key.
        return [
            ssm_key["TargetKeyId"]
            for ssm_key in boto3.client("kms").list_aliases()["Aliases"]
            if self.kms_key_alias in ssm_key["AliasName"]
        ][0]

    @property
    def repository_url(self) -> str:
        return f"{self.account_id}.dkr.ecr.{self.region}.amazonaws.com/{self.repository_name}"

    @property
    def registry_url(self) -> str:
        return f"{self.account_id}.dkr.ecr.{self.region}.amazonaws.com"



from sentential.cli.root import root
from sentential.cli.args import args
from sentential.cli.envs import envs
from sentential.cli.configs import configs
from sentential.cli.deploy import deploy
from sentential.cli.destroy import destroy
from sentential.cli.logs import logs
from sentential.lib.exceptions import SntlException, DockerException, AWS_EXCEPTIONS

root.add_typer(args, name="args", help="build arguments")
root.add_typer(envs, name="envs", help="environment variables")
root.add_typer(configs, name="configs", help="provisioning")
root.add_typer(deploy, name="deploy", help="create deployment")
root.add_typer(destroy, name="destroy", help="destroy deployment")
root.add_typer(logs, name="logs", help="logging")


def main():
    try:
        root()
    except SntlException as e:
        print(f"SENTENTIAL: {e}")
        exit(1)
    except AWS_EXCEPTIONS as e:
        print(f"AWS: {e}")
        exit(1)
    except DockerException as e:
        print(f"DOCKER: {e}")
        exit(1)



from datetime import datetime
from enum import Enum
from lib2to3.pgen2.token import OP
from pathlib import PosixPath
from typing import List, Union, Optional, Dict
from pydantic import BaseModel, Field, validator
from sentential.support.shaper import Shaper

#
# Global Constants
#
CURRENT_WORKING_IMAGE_TAG = "cwi"

#
# Internally Defined Shapes
#
class Image(BaseModel):
    id: str
    digest: Union[str, None]
    tags: List[str]
    versions: List[str]
    # arch: str TODO: https://github.com/gabrieldemarmiesse/python-on-whales/pull/378

    @validator("versions")
    def coerce_versions(cls, v):
        if v is not None:
            uniq = list(set(v))
            return uniq
        else:
            return []

    @validator("tags")
    def coerce_tags(cls, v):
        if v is not None:
            uniq = list(set(v))
            return uniq
        else:
            return []


class ImageView(Image):
    href: List[str] = []

    @validator("href")
    def uniq(cls, v):
        return list(set(v))

    @validator("id")
    def humanize_id(cls, v):
        return v.replace("sha256:", "")[0:12]

    @validator("digest")
    def humanize_digest(cls, v):
        if v:
            return v.replace("sha256:", "")[0:12]


class Function(BaseModel):
    image: Image
    name: str
    role_arn: str
    role_name: str
    region: str
    arn: str
    web_console_url: Union[str, None]
    public_url: Union[str, None]


class Provision(Shaper):
    storage: int = Field(default=512, description="ephemeral storage (mb)")
    memory: int = Field(default=128, description="allocated memory (mb)")
    timeout: int = Field(default=3, description="timeout (s)")
    subnet_ids: List[str] = Field(default=[], description="subnet ids")
    security_group_ids: List[str] = Field(default=[], description="security group ids")
    auth_type: str = Field(default="NONE", description="auth type (--public-url)")
    allow_headers: List[str] = Field(
        default=["*"], description="CORS AllowHeaders (--public-url)"
    )
    allow_methods: List[str] = Field(
        default=["*"], description="CORS AllowMethods (--public-url)"
    )
    allow_origins: List[str] = Field(
        default=["*"], description="CORS AllowOrigins (--public-url)"
    )
    expose_headers: List[str] = Field(
        default=["*"], description="CORS ExposeHeaders (--public-url)"
    )

    @validator("auth_type")
    def is_valid_auth_type(cls, v):
        valid_auth_types = ["NONE", "AWS"]
        if v not in valid_auth_types:
            raise ValueError(f"auth_type must be one of {', '.join(valid_auth_types)}")
        return v


#
# IAM
#
class AWSPolicyStatement(BaseModel):
    """https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_statement.html"""

    Effect: str
    Action: Union[str, List[str]]
    Principal: Optional[dict]
    Resource: Optional[Union[str, List[str]]]
    Condition: Optional[dict]


class AWSPolicyDocument(BaseModel):
    """https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html"""

    Version: str = "2012-10-17"
    Statement: List[AWSPolicyStatement]


class AWSCallerIdentity(BaseModel):
    UserId: str
    Account: str
    Arn: str


#
# Lambda
#
class Runtimes(Enum):
    """https://gallery.ecr.aws/lambda?page=1"""

    python = "python"
    dotnet = "dotnet"
    java = "java"
    go = "go"
    nodejs = "nodejs"
    provided = "provided"
    ruby = "ruby"


LAMBDA_ROLE_POLICY_JSON = (
    AWSPolicyDocument(
        Statement=[
            AWSPolicyStatement(
                Effect="Allow",
                Principal={"Service": "lambda.amazonaws.com"},
                Action="sts:AssumeRole",
            )  # type: ignore
        ]
    )
).json(exclude_none=True)


class AwsFunctionVpcConfig(BaseModel):
    SubnetIds: List[str]
    SecurityGroupIds: List[str]
    VpcId: str


class AwsFunctionEnvironment(BaseModel):
    Variables: Dict[str, str]


class AwsFunctionEphemeralStorage(BaseModel):
    Size: int


class AwsFunctionConfiguration(BaseModel):
    FunctionName: str
    FunctionArn: str
    Role: str
    CodeSize: int
    Description: str
    Timeout: int
    MemorySize: int
    LastModified: datetime
    CodeSha256: str
    Version: str
    VpcConfig: AwsFunctionVpcConfig
    Environment: AwsFunctionEnvironment
    TracingConfig: Dict[str, str]
    RevisionId: str
    State: str
    StateReason: Optional[str]
    StateReasonCode: Optional[str]
    PackageType: str
    Architectures: List[str]
    EphemeralStorage: AwsFunctionEphemeralStorage


class AwsFunctionCode(BaseModel):
    RepositoryType: str
    ImageUri: str
    ResolvedImageUri: str


class AwsFunctionCors(BaseModel):
    AllowHeaders: List[str]
    AllowMethods: List[str]
    AllowOrigins: List[str]
    ExposeHeaders: List[str]


class AwsFunctionPublicUrl(BaseModel):
    FunctionUrl: str
    FunctionArn: str
    AuthType: str
    Cors: AwsFunctionCors
    CreationTime: datetime
    LastModifiedTime: datetime


class AwsFunction(BaseModel):
    Configuration: AwsFunctionConfiguration
    Code: AwsFunctionCode


#
# Pathing
#
class Paths(BaseModel):
    root: PosixPath
    sntl: PosixPath
    src: PosixPath
    dockerfile: PosixPath
    wrapper: PosixPath
    policy: PosixPath
    shapes: PosixPath


def derive_paths(root: PosixPath = PosixPath(".")):
    return Paths(
        root=root,
        sntl=PosixPath(f"{root}/.sntl"),
        src=PosixPath(f"{root}/src"),
        dockerfile=PosixPath(f"{root}/Dockerfile"),
        wrapper=PosixPath(f"{root}/.sntl/wrapper.sh"),
        policy=PosixPath(f"{root}/policy.json"),
        shapes=PosixPath(f"{root}/shapes.py"),
    )


#
# API Gateway
#
class ApiGatewayIntegration(BaseModel):
    ConnectionType: str = "INTERNET"
    Description: str = "managed by sentential"
    IntegrationId: Optional[str] = None
    IntegrationMethod: str = "ANY"
    IntegrationType: str = "AWS_PROXY"
    IntegrationUri: str
    PayloadFormatVersion: str = "2.0"
    TimeoutInMillis: int = 30000
    RequestParameters: Optional[Dict[str, str]] = {}


class ApiGatewayRoute(BaseModel):
    ApiKeyRequired: bool
    AuthorizationType: str
    RouteId: str
    RouteKey: str
    Target: Optional[str]
    Integration: Optional[ApiGatewayIntegration]


class ApiGatewayMapping(BaseModel):
    ApiId: str
    ApiMappingId: str
    ApiMappingKey: str
    Stage: str
    Routes: List[ApiGatewayRoute] = []


class ApiGatewayDomain(BaseModel):
    DomainName: str
    DomainNameConfigurations: Optional[List[Dict]]
    Tags: Dict[str, str] = {}
    Mappings: List[ApiGatewayMapping] = []


class ApiGatewayParsedUrl(BaseModel):
    ApiId: str
    ApiMappingId: str
    ApiMappingKey: str
    RouteKey: str
    RouteId: Optional[str]
    Verb: str = "Any"
    Route: str
    FullPath: str


class LambdaPermission(BaseModel):
    FunctionName: str
    StatementId: str
    Action: str
    Principal: str
    SourceArn: str
    SourceAccount: str
    EventSourceToken: str
    Qualifier: str
    RevisionId: str
    PrincipalOrgID: str
    FunctionUrlAuthType: str = "None"



from sentential.lib.exceptions import ApiGatewayResourceNotFound
from sentential.lib.clients import clients
from sentential.lib.ontology import Ontology
from sentential.lib.shapes import (
    Function,
    ApiGatewayDomain,
    ApiGatewayIntegration,
    ApiGatewayMapping,
    ApiGatewayParsedUrl,
    ApiGatewayRoute,
)

from furl import furl
from typing import List
from urllib.parse import urlparse


def normalize_url(url: str) -> str:
    normalized_url = "/".join(furl(url).path.normalize().segments)
    return normalized_url


def normalize_route(url: str) -> str:
    normalized_route = furl(url).path.normalize()
    decoded = "/".join(normalized_route.segments)
    return f"/{decoded}"


class AwsApiGatewayDriver:
    def __init__(self, ontology: Ontology) -> None:
        self.ontology = ontology
        self.context = self.ontology.context
        self.account_id = self.context.account_id
        self.partition = self.context.partition
        self.region = self.context.region
        self.repo_name = self.context.repository_name
        self.statement_id = f"{self.partition}-{self.region}-{self.repo_name}"

    @classmethod
    def sentential_domains(cls) -> List[ApiGatewayDomain]:
        sentential_domains = []
        for domain in clients.api_gw.get_domain_names()["Items"]:
            domain = ApiGatewayDomain(**domain)
            if "sentential" in domain.Tags.keys():
                sentential_domains.append(domain)
        return sentential_domains

    @classmethod
    def domains(cls) -> List[ApiGatewayDomain]:
        domains = cls.sentential_domains()
        for domain in domains:
            for mapping in clients.api_gw.get_api_mappings(
                DomainName=domain.DomainName
            )["Items"]:
                mapping = ApiGatewayMapping(**mapping)
                domain.Mappings.append(mapping)
                for route in clients.api_gw.get_routes(ApiId=mapping.ApiId)["Items"]:
                    route = ApiGatewayRoute(**route)
                    mapping.Routes.append(route)
                    if route.Target:
                        integration_id = route.Target.split("/")[-1]
                        integration = clients.api_gw.get_integration(
                            ApiId=mapping.ApiId, IntegrationId=integration_id
                        )
                        route.Integration = ApiGatewayIntegration(**integration)
        return domains

    @classmethod
    def autocomplete(cls) -> List[str]:
        urls = []
        for domain in cls.domains():
            urls.append(normalize_url(domain.DomainName))
            for mapping in domain.Mappings:
                for route in mapping.Routes:
                    uri = route.RouteKey.split(" ")[-1]
                    full_path = normalize_url(
                        f"{domain.DomainName}/{mapping.ApiMappingKey}/{uri}"
                    )
                    urls.append(str(full_path))
        if len(urls) == 0:
            return ["no discoverable api gateway endpoints (just link to docs)"]
        return list(set(urls))

    def parse(self, url: str) -> ApiGatewayParsedUrl:
        parsed = urlparse(f"https://{url}")
        domains = self.domains()
        for domain in domains:
            if domain.DomainName == parsed.netloc:
                # if matching route exists under mapping, return it
                for mapping in domain.Mappings:
                    for route in mapping.Routes:
                        existant_verb, existant_mount = route.RouteKey.split(" ")
                        found_path = normalize_route(
                            f"/{mapping.ApiMappingKey}/{existant_mount}"
                        )
                        if parsed.path == found_path:
                            full_path = normalize_url(
                                f"{domain.DomainName}/{found_path}"
                            )
                            return ApiGatewayParsedUrl(
                                ApiId=mapping.ApiId,
                                ApiMappingId=mapping.ApiMappingId,
                                ApiMappingKey=mapping.ApiMappingKey,
                                RouteId=route.RouteId,
                                RouteKey=route.RouteKey,
                                Route=str(
                                    furl(existant_mount)
                                    .path.remove("/{proxy+}")
                                    .remove("/{proxy}")
                                ),
                                Verb=existant_verb,
                                FullPath=str(full_path),
                            )

                # else return new path under matching mapping
                # sort so that "startswith" conditional matches against longest to shortest possibilities
                domain.Mappings.sort(key=lambda m: len(m.ApiMappingKey), reverse=True)
                for mapping in domain.Mappings:
                    map_key_path = f"/{mapping.ApiMappingKey}"
                    if parsed.path.startswith(map_key_path):
                        full_path = normalize_url(f"{domain.DomainName}/{parsed.path}")
                        route_key = normalize_route(
                            parsed.path.replace(map_key_path, "", 1)
                        )
                        return ApiGatewayParsedUrl(
                            ApiId=mapping.ApiId,
                            ApiMappingId=mapping.ApiMappingId,
                            ApiMappingKey=mapping.ApiMappingKey,
                            RouteId=None,
                            RouteKey=f"ANY {route_key}",
                            Route=str(
                                furl(route_key)
                                .path.remove("/{proxy+}")
                                .remove("/{proxy}")
                            ),
                            FullPath=str(full_path),
                        )

        # else explode
        raise ApiGatewayResourceNotFound(
            f"invalid url ({url}): bad formatting, or attempting to mount to non-existant domain/mapping"
        )

    def _ensure_integration(
        self, parsed_url: ApiGatewayParsedUrl, function: Function
    ) -> ApiGatewayIntegration:
        desired_integration = ApiGatewayIntegration(
            IntegrationUri=function.arn,
            RequestParameters={
                "overwrite:path": "/$request.path.proxy",
                "overwrite:header.X-Forwarded-Prefix": parsed_url.Route,
            },
        )

        # if the desired integration already exists, return it
        for integration in clients.api_gw.get_integrations(ApiId=parsed_url.ApiId)[
            "Items"
        ]:
            current_integration = ApiGatewayIntegration(**integration)
            if current_integration.IntegrationId is not None:
                current_integration.IntegrationId = None
                if desired_integration == current_integration:
                    response = clients.api_gw.get_integration(
                        ApiId=parsed_url.ApiId,
                        IntegrationId=integration["IntegrationId"],
                    )
                    return ApiGatewayIntegration(**response)

        # else create the desired integration and return it
        desired_integration = desired_integration.dict(exclude={"IntegrationId"})
        desired_integration["ApiId"] = parsed_url.ApiId
        response = clients.api_gw.create_integration(**desired_integration)
        return ApiGatewayIntegration(**response)

    def _ensure_route(
        self, parsed_url: ApiGatewayParsedUrl, integration: ApiGatewayIntegration
    ) -> ApiGatewayRoute:
        target = f"integrations/{integration.IntegrationId}"
        try:
            response = clients.api_gw.create_route(
                ApiId=parsed_url.ApiId,
                RouteKey=parsed_url.RouteKey,
                Target=target,
            )
        except clients.api_gw.exceptions.ConflictException:
            response = clients.api_gw.update_route(
                ApiId=parsed_url.ApiId,
                RouteId=parsed_url.RouteId,
                RouteKey=parsed_url.RouteKey,
                Target=target,
            )
        return ApiGatewayRoute(**response)

    def _ensure_permission(self, parsed_url: ApiGatewayParsedUrl, function: Function):
        verb, route = parsed_url.RouteKey.split(" ")
        try:
            clients.lmb.remove_permission(
                FunctionName=function.name,
                StatementId=self.statement_id,
            )
        except clients.lmb.exceptions.ResourceNotFoundException:
            pass

        response = clients.lmb.add_permission(
            FunctionName=function.arn,
            StatementId=self.statement_id,
            Action="lambda:InvokeFunction",
            Principal="apigateway.amazonaws.com",
            SourceArn=f"arn:aws:execute-api:{self.region}:{self.account_id}:{parsed_url.ApiId}/*/*/{{proxy+}}",
        )
        return response

    def put_route(self, url: str, function: Function) -> ApiGatewayRoute:
        parsed_url = self.parse(url)
        integration = self._ensure_integration(parsed_url, function)
        route = self._ensure_route(parsed_url, integration)
        permission = self._ensure_permission(parsed_url, function)
        return route

    def delete_routes(self, function: Function) -> None:
        domains = self.domains()
        to_delete = []
        for domain in domains:
            for mapping in domain.Mappings:
                for route in mapping.Routes:
                    if route.Integration:
                        if route.Integration.IntegrationUri == function.arn:
                            to_delete.append([mapping, route])

        for mapping, route in to_delete:
            try:
                clients.api_gw.delete_route(ApiId=mapping.ApiId, RouteId=route.RouteId)
            except clients.api_gw.exceptions.NotFoundException:
                pass

        for mapping, route in to_delete:
            try:
                clients.api_gw.delete_integration(
                    ApiId=mapping.ApiId,
                    IntegrationId=route.Integration.IntegrationId,
                )
            except clients.api_gw.exceptions.NotFoundException:
                pass



import typer
from sentential.lib.clients import clients
from sentential.lib.template import Init
from sentential.lib.shapes import Runtimes
from sentential.lib.drivers.local_lambda import LocalLambdaDriver
from sentential.lib.drivers.aws_lambda import AwsLambdaDriver
from sentential.lib.semver import SemVer
from sentential.lib.ontology import Ontology
from sentential.lib.joinery import Joinery
from sentential.lib.shapes import CURRENT_WORKING_IMAGE_TAG

from rich import print

root = typer.Typer()


@root.command()
def init(repository_name: str, runtime: Runtimes):
    """initialize sentential project"""
    aws_supported_image = f"public.ecr.aws/lambda/{runtime.value}:latest"
    Init(repository_name, aws_supported_image).scaffold()


@root.command()
def build():
    """build lambda image"""
    local = LocalLambdaDriver(Ontology())
    print(local.build(CURRENT_WORKING_IMAGE_TAG))


@root.command()
def publish(major: bool = typer.Option(False), minor: bool = typer.Option(False)):
    """publish lambda image"""
    ontology = Ontology()
    aws = AwsLambdaDriver(ontology)
    local = LocalLambdaDriver(ontology)
    version = SemVer(aws.images()).next(major, minor)
    print(local.publish(CURRENT_WORKING_IMAGE_TAG, version))


@root.command()
def login():
    """login to ecr"""
    clients.docker.login_ecr()


@root.command()
def ls():
    """list image information"""
    print(Joinery(Ontology()).list(["tags"]))


@root.command()
def clean(remote: bool = typer.Option(False)):
    """clean images"""
    ontology = Ontology()
    aws = AwsLambdaDriver(ontology)
    local = LocalLambdaDriver(ontology)

    for image in local.images():
        clients.docker.image.remove(image.id, force=True)

    if remote:
        images = []
        for image in aws.images():
            images.append({"imageDigest": image.digest})

        clients.ecr.batch_delete_image(
            repositoryName=ontology.context.repository_name, imageIds=images
        )
